#! /bin/sh -efu
# $Id: pkg_replace.sh,v 1.7 2006/10/15 19:19:17 securedog Exp $

usage() {
	echo "usage: ${0##*/} [-habBcCiInqrRuvwW] [-l file] [-m args]"
	echo "         [[pkgname[=package]] [package] ...]"
	exit ${1:-0}
}

is_yes() {
	case ${1-} in
	[Yy][Ee][Ss]|[Yy]|[Tt][Rr][Uu][Ee])
		return 0 ;;
	*)	return 1 ;;
	esac
}

warn() {
	echo "** ${@-}" >&2
}

info() {
	echo "--->  ${@-}"
}

prompt_yesno() {
	local prompt default input

	prompt=${1-"OK?"}
	default=${2-"yes"}

	echo -n "${prompt} [${default}] " >&2
	read input

	is_yes ${input:-"${default}"}
}

init_options() {
	opt_all=NO
	opt_backup=YES
	opt_keep_backup=NO
	opt_beforeclean=NO
	opt_afterclean=YES
	opt_force=NO
	opt_interactive=NO
	opt_use_index=NO
	opt_log_output=
	opt_make_args=
	opt_noexecute=NO
	opt_noconf=NO
	opt_depends=NO
	opt_required_by=NO
	opt_preserve_libs=YES
	opt_verbose=NO
	opt_debug=0
	opt_replace=
	do_replace=
	log_output=
	log_format="+:done -:ignored *:skipped !:failed"
#ifdef WITH_EXPERIMENTAL
	MAKE_ARGS=
	BEFOREINSTALL=
	BEFOREDEINSTALL=
	AFTERINSTALL=
#endif
}

init_variables() {
#ifdef WITH_PORTS
	: ${PORTSDIR="/usr/ports"}
	: ${PACKAGES="${PORTSDIR}/packages"}
#endif
#ifdef WITH_PKGSRC
	: ${PKGSRCDIR="/usr/pkgsrc"}
	: ${PACKAGES="${PKGSRCDIR}/packages"}
#endif
	: ${PKGREPOSITORY="${PACKAGES}/All"}
	: ${PKG_DBDIR="/var/db/pkg"}
	: ${PKG_TMPDIR=${TMPDIR:-"/var/tmp"}}
#ifdef WITH_PORTS
	: ${PKG_SUFX=".tbz"}
#endif
#ifdef WITH_PKGSRC
	: ${PKG_SUFX=".tgz"}
#endif
	: ${PKG_BACKUP_DIR=${PKGREPOSITORY}}
	: ${PKGCOMPATDIR="/usr/local/lib/compat/pkg"}
	: ${MAKE="make"}
	cwd=${PWD}
#ifdef WITH_PORTS
	export PORTSDIR \
#endif
#ifdef WITH_PKGSRC
	PORTSDIR=${PKGSRCDIR}
	export PKGSRCDIR \
#endif
	       PACKAGES PKGREPOSITORY PKG_DBDIR PKG_TMPDIR PKG_SUFX PKGCOMPATDIR
}

compat_variables() {
	if [ -n "${PKG_REPLACE-}" ]; then
		parse_options ${PKG_REPLACE-}
	fi
}

parse_options() {
	local OPT OPTARG OPTIND

	while getopts habBcCfiIl:m:nqrRuvwWX OPT; do
		case ${OPT} in
		a)	opt_all=YES ;;
		b)	opt_keep_backup=YES ;;
		B)	opt_backup=NO ;;
		c)	opt_beforeclean=YES ;;
		C)	opt_afterclean=YES ;;
		h)	usage ;;
		f)	opt_force=YES ;;
		i)	opt_interactive=YES ;;
		I)	opt_use_index=YES ;;
		l)	opt_log_output=${OPTARG} ;;
		m)	opt_make_args="${opt_make_args} ${OPTARG}" ;;
		n)	opt_noexecute=YES ;;
		q)	opt_noconf=YES ;;
		r)	opt_required_by=YES ;;
		R)	opt_depends=YES ;;
		u)	opt_preserve_libs=NO ;;
		v)	opt_verbose=YES ;;
		w)	opt_beforeclean=NO ;;
		W)	opt_afterclean=NO ;;
		X)	opt_debug=$((${opt_debug}+1)) ;;
		\?)	usage 1 ;;
		esac
	done

	OPTC=$((${OPTIND}-1))
}

#ifdef WITH_EXPERIMENTAL
parse_config() {
	local LINE q lineno var val open_bracket hash_key hash_val

	[ -r "$1" ] || return 0

	q="['\"]"
	lineno=0
	open_bracket=

	while read LINE; do
		lineno=$((${lineno}+1))

		case ${LINE} in
		''|\#*) continue ;;
		esac

		if [ -n "${open_bracket}" ]; then
			case ${LINE} in
			*$q?*$q*=\>*$q?*$q*) ;;
			*[\)}]) open_bracket=; continue ;;
			*)
				warn "Syntax error at line ${lineno}: ${LINE}"
				return 1 ;;
			esac

			hash_key=${LINE#*$q}
			hash_key=${hash_key%%$q*}
			hash_val=${LINE#*=>*$q}
			hash_val=${hash_val%$q*}

			eval ${var}=\"\${${var}:+\$${var}'
'}${hash_key}=\${hash_val}\"
			continue
		fi

		case ${LINE} in
		?*=*)	var=${LINE%%=*}; val=${LINE#*=} ;;
		*)	var="syntax-error" ;;
		esac

		case ${var} in
		*[!0-9A-Za-z_]*|[0-9]*)
			warn "Syntax error at line ${lineno}: ${LINE}"
			return 1 ;;
		esac

		case ${val} in
		*[\({])	open_bracket=1 ;;
		*)	eval ${var}=${val} ;;
		esac
	done < "$1"
}

parse_exec_script() {
	local IFS hash key command

	IFS='
'
	eval hash=\${$1-}
	for key in ${hash}; do
		case ${pkg_origin} in
		${key%%=*})
			command=${key#*=}
			info "Executing a $1 command: ${command}"
			( eval "${command}" ) || :
		esac
	done
}
#endif

set_port_name() {
	port_name=
	if is_yes ${opt_use_index}; then
		port_name=$(grep --mmap -m 1 -o "^[^|]*|[^|]*${1#${PORTSDIR}/}|" "${PORTSDIR}/INDEX")
		port_name="${port_name%%|*}"
	elif [ -e "$1/Makefile" ] && cd "$1"; then
#ifdef WITH_PORTS
#ifdef WITH_OPENBSD
		port_name=$(${MAKE} show=FULLPKGNAME\${SUBPACKAGE})
#else
		port_name=$(${MAKE} -VPKGNAME)
#endif
#endif
#ifdef WITH_PKGSRC
		port_name=$(${MAKE} -V '${PKGNAME}')
#endif
	fi

	if [ -z "${port_name}" ]; then
		return 1
	fi
}

set_pkg_name() {
	pkg_name=$1
}

set_pkg_origin() {
	local LINE

	pkg_origin=
#ifdef WITH_PORTS
	if [ -r "${PKG_DBDIR}/$1/+CONTENTS" ]; then
#endif
#ifdef WITH_PKGSRC
	if [ -r "${PKG_DBDIR}/$1/+BUILD_INFO" ]; then
#endif
		while read LINE; do
			case ${LINE} in
#ifdef WITH_PORTS
			@comment\ ORIGIN:*)
				pkg_origin="${LINE#@comment ORIGIN:}"
				break ;;
			[!@]*)	break ;;
#endif
#ifdef WITH_PKGSRC
			PKGPATH=*)
				pkg_origin="${LINE#PKGPATH=}"
				break ;;
#endif
			esac
#ifdef WITH_PORTS
		done < "${PKG_DBDIR}/$1/+CONTENTS"
#endif
#ifdef WITH_PKGSRC
		done < "${PKG_DBDIR}/$1/+BUILD_INFO"
#endif
	fi
}

set_pkg_pkgdir() {
	pkg_pkgdir=${PKG_DBDIR}/$1
}

set_pkg_portdir() {
	pkg_portdir=${PORTSDIR}/$1
}

correct_pkg() {
	if [ -z "${pkg_origin}" ]; then
		if [ -d "${pkg_pkgdir}" ]; then
			warn "'${pkg_name}' has no origin recorded."
		else
			warn "'${pkg_name}' is not installed. (broken dependencies?)"
		fi
		return 1
	elif [ ! -d "${pkg_portdir}" ]; then
		warn "${pkg_name} (${pkg_origin}): perhaps moved or obsoleted."
		trace_moved "${pkg_origin}" || return 1
	fi
}

pkg_glob() {
	local - pattern name contents

	set +f
	while [ $# -gt 0 ]; do
		pattern=${1#${PKG_DBDIR}/}

		case ${pattern} in
		''|*[\<\>/]*)	shift; continue ;;
		*\**|*-pl[0-9]*|*-[0-9]*[0-9.][a-z]|*-[0-9]*[0-9]) ;;
		*)		pattern="${pattern}-[0-9]*[0-9a-z]" ;;
		esac

		for contents in ${PKG_DBDIR}/${pattern}/+CONTENTS; do
			if [ -e "${contents}" ]; then
				name=${contents%/+CONTENTS}

				if is_yes ${opt_depends}; then
					pkg_depends "${name##*/}"
				fi

				echo "${name##*/}"

				if is_yes ${opt_required_by}; then
					pkg_required_by "${name##*/}"
				fi
			else
				warn "No such installed package: $1"
			fi
		done
		shift
	done
}

pkg_depends() {
	local dep pkg xvar

	case ${2-} in
	\$*)	xvar=${2#$}; eval ${xvar}= ;;
	*)	xvar= ;;
	esac

	if [ -r "${PKG_DBDIR}/$1/+CONTENTS" ]; then
		while read dep; do
			case ${dep} in
#ifdef WITH_PORTS
			@pkgdep\ *)
				if [ -n "${xvar}" ]; then
					eval ${xvar}=\"\$${xvar} ${dep#@pkgdep }\"
				else
					echo "${dep#@pkgdep }"
				fi ;;
#endif
#ifdef WITH_PKGSRC
			@pkgdep\ *)
				case ${dep#@pkgdep } in
				*[\<\>{}*]*)	pkg=$(pkg_info -e "${dep#@pkgdep }") ;;
				*)		pkg=${dep#@pkgdep } ;;
				esac

				if [ -n "${xvar}" ]; then
					eval ${xvar}=\"\$${xvar} ${pkg}\"
				else
					echo "${pkg}"
				fi ;;
#endif
			[!@]*)	break ;;
			esac

		done < "${PKG_DBDIR}/$1/+CONTENTS"
	fi
}

pkg_required_by() {
	local req xvar
	
	case ${2-} in
	\$*)	xvar=${2#$}; eval ${xvar}= ;;
	*)	xvar= ;;
	esac

	if [ -r "${PKG_DBDIR}/$1/+REQUIRED_BY" ]; then
		while read req; do
			if [ -n "${xvar}" ]; then
				eval ${xvar}=\"\$${xvar} ${req}\"
			else
				echo "${req}"
			fi
		done < "${PKG_DBDIR}/$1/+REQUIRED_BY"
	fi
}

pkg_sort() {
	local leaves checked pkg_node pkg depends

	leaves=
	checked=
	pkg_node=" $@ "

	while [ $# -gt 0 ]; do
		case ${checked}${leaves} in
		*" $1 "*) shift; continue ;;
		esac

		if [ $# -gt 1 ] && [ -s "${PKG_DBDIR}/$1/+REQUIRED_BY" ]; then
			pkg_depends "$1" '$depends'

			for pkg in ${depends}; do
				case ${checked} in
				*" ${pkg} "*) continue ;;
				esac
				case ${pkg_node} in
				*" ${pkg} "*) checked="${checked:- }${pkg} " ;;
				esac
			done

			checked="${checked:- }$1 "
		else
			leaves="${leaves:- }$1 "
		fi
		shift
	done

	echo ${checked} ${leaves}
}

#ifdef WITH_PORTS
guess_up_to_date() {
	local IFS all pkg

	all=" $@ "
	do_replace=
	IFS='
'
	for pkg in $(pkg_version -qvIl\< "${PORTSDIR}/INDEX"); do
		case ${all} in
		*" ${pkg%% *} "*) do_replace="${do_replace} ${pkg%% *}" ;;
		esac
	done
}
#endif

try() {
	local exit

	if [ ${opt_debug} -ge 2 ]; then
		echo "[$@]"
	elif "$@"; exit=$?; [ ${exit} != 0 ]; then
		warn "Command failed (exit code ${exit}): $@"
		return ${exit}
	fi
}

build_port() {
	info "Building '$1'"
#ifdef WITH_EXPERIMENTAL
	parse_exec_script BEFOREBUILD
#endif

	if is_yes ${opt_beforeclean}; then
		clean_port "$1" || return 1
	fi

	if ! { cd "$1" && try ${MAKE_CMD}; }; then
		return 1
	fi
}

install_port() {
	local install_args

	install_args=
	is_yes ${opt_force} && install_args="-DFORCE_PKG_REGISTER"

	info "Installing '$1'"

	if ! { cd "$1" && try ${MAKE_CMD} ${install_args} install; }; then
		return 1
	fi

	if is_yes ${opt_afterclean}; then
		clean_port "$1" || return 1
	fi
#ifdef WITH_EXPERIMENTAL

	parse_exec_script AFTERINSTALL
#endif
}

install_package() {
	local install_args

	install_args=
	is_yes ${opt_force} && install_args="-f"
	is_yes ${opt_verbose} && install_args="${install_args} -v"

	info "Installing '$1'"

	if ! try pkg_add ${install_args} "$1"; then
		return 1
	fi
#ifdef WITH_EXPERIMENTAL

	parse_exec_script AFTERINSTALL
#endif
}

deinstall_package() {
	local deinstall_args

	deinstall_args=
	is_yes ${opt_force} && deinstall_args="-f"
	is_yes ${opt_verbose} && deinstall_args="${deinstall_args} -v"

	info "Deinstalling '$1'"
#ifdef WITH_EXPERIMENTAL
	parse_exec_script BEFOREDEINSTALL
#endif

	if [ ${opt_debug} -lt 2 ] && [ ! -w "${PKG_DBDIR}" ]; then
		warn "You do not own ${PKG_DBDIR}."
		return 1
	fi

	if ! try pkg_delete ${deinstall_args} "$1"; then
		return 1
	fi
}

clean_port() {
	local clean_args

#ifdef WITH_PORTS
	clean_args=
#endif
#ifdef WITH_PKGSRC
	clean_args="CLEANDEPENDS=yes"
#endif

	info "Cleaning '$1'"

	if ! { cd "$1" && try ${MAKE_CMD} ${clean_args} clean; }; then
		return 1
	fi
}

#ifdef WITH_PKGSRC
pkg_backup() {
	local pkg_pkgdir package plist create_args arg

	set_pkg_pkgdir "$1"
	package=$2
	create_args=

	if ! { cd "${pkg_pkgdir}" && file_exist "+CONTENTS" && \
		plist=$(mktemp -t "$1"); }; then
		return 1
	fi

	sed  -e '/^@blddep /d' -e '/^@comment [^[:space:]]*:/d' \
	     -e '/^@name /d' -e '/^@mtree /d' \
	     -e '/^@pkgcfl /d' -e '/^@pkgdep /d' \
		"+CONTENTS" > "${plist}" || return 1

	while read LINE; do
		case ${LINE} in
		@pkgcfl\ *)	create_args="${create_args} -C ${LINE#@pkgcfl }" ;;
		@pkgdep\ *)	create_args="${create_args} -P ${LINE#@pkgdep }" ;;
		[!@]*)	break ;;
		esac
	done < "+CONTENTS"

	for arg in \
		b:BUILD_VERSION B:BUILD_INFO c:COMMENT d:DESC \
		D:DISPLAY i:INSTALL k:DEINSTALL m:MTREE n:PRESERVE \
		r:REQUIRE s:SIZE_PKG S:SIZE_ALL; do

		if [ -e "+${arg#*:}" ]; then
			create_args="${create_args} -${arg%%:*} +${arg#*:}"
		fi
	done

	try pkg_create ${create_args} -lf "${plist}" "${package}" || {
		try rm "${plist}"; return 1; }

	try rm "${plist}" && cd "${cwd}"
}
#endif

create_old_package() {
	if is_yes ${opt_backup}; then
		info "Backing up the old version"
#ifdef WITH_PORTS
		try pkg_create -b "$1" "$2" || return 1
#endif
#ifdef WITH_PKGSRC
		try pkg_backup "$1" "$2" || return 1
#endif
	fi
}

create_old_required_by() {
	if [ -e "$1" ]; then
		info "Backing up the +REQUIRED_BY file"
		try cp -f "$1" "$2" || return 1
	fi
}

restore_package() {
	if [ -e "$1" ]; then
		info "Restoring the old version"
		opt_force=YES install_package "$1" || return 1
	fi
}

restore_required_by() {
	if [ -e "$1" ] && [ ! -e "$2" ]; then
		info "Restoring the +REQUIRED_BY file"
		try mv -f "$1" "$2" || return 1
	fi
}

process_old_package() {
	if is_yes ${opt_keep_backup} && [ -e "$1" ] && [ ! -e "${PKG_BACKUP_DIR}/${1##*/}" ]; then
		info "Keeping old package in '${PKG_BACKUP_DIR}'"
		make_dirs "${PKG_BACKUP_DIR}" || return 1
		try mv -f "$1" "${PKG_BACKUP_DIR}" || return 1
	fi
}

cleanup() {
	try rm -rf "$@" || warn "Couldn't remove '$@'."
}

trace_moved() {
#ifdef WITH_PORTS
	local IFS msg checked moved origin

	file_exist "${PORTSDIR}/MOVED" || return 1

	msg=
	origin=$1
	checked=
	IFS='|'

	while moved=$(grep -m1 "^$1|" "${PORTSDIR}/MOVED"); do
		set -- ${moved}

		if [ -n "$2" ]; then
			checked="${checked:-|}$1|"

			case ${checked} in
			*"|${2}|"*)
				warn "Failed. (${msg} -> ???)"
				break ;;
			esac

			msg="${msg:+${msg} }-> $2"
			set_pkg_portdir "$2"

			if [ -d "${pkg_portdir}" ]; then
				warn "'${origin}' has been moved. (${msg})"
				pkg_origin=$2
				return 0
			fi

			set -- "$2"
		else
			warn "'$1' has been removed from ports tree:"
			warn "\"$4\" ($3)"
			return 1
		fi
	done

	warn "Port directory not found: ${origin}"
	return 1
#endif
#ifdef WITH_PKGSRC
	warn "See ${PORTSDIR}/doc/CHANGES-*."
	return 1
#endif
}

set_package_name() {
	local opt

	opt=
	case $1 in
	*.gz|*.tgz)	opt=z ;;
	*.bz2|*.tbz)	opt=j ;;
	esac

	port_name=`tar x${opt}Of "$1" "+CONTENTS" |
	while read LINE; do
		case ${LINE} in
		@name\ *)
			echo "${LINE#@name }"
			break ;;
		[!@]*)  break ;;
		esac
	done`

	if [ -z "${port_name}" ]; then
		warn "Unable to open contents file: ${file} (not a package?)"
		return 1
	fi
}

perform() {
	local MAKE_CMD pkg_name pkg_origin pkg_pkgdir pkg_portdir port_name use_package status log

	while [ $# -gt 0 ]; do
		log=
		status=
		set_pkg_name "$1"
		set_pkg_pkgdir "$1"
		shift

		if perform_preprocess; then
			if ! perform_replace; then
				status=failed
				warn "Fix the problem and try again."
				echo
			fi
		else
			status=skipped
			warn "Skipping '${pkg_name}'."
		fi

		add_log "${status:-ignored}" "${pkg_name}" ${log:+" (${log})"}
	done
}

perform_preprocess() {
	local IFS i replace_with opt_make_args

	use_package=
	replace_with=

	for i in ${opt_replace}; do
		case ${pkg_name} in
		${i%=*}|${i%=*}-[0-9]*)
			replace_with="${i##*=}"
			break ;;
		esac
	done

	set_pkg_origin "${pkg_name}" 
	case ${replace_with} in
	*.t[bg]z)	use_package=${replace_with} ;;
	?*)		pkg_origin=${replace_with#${PORTSDIR}/} ;;
	esac

	if [ -n "${use_package}" ]; then
		set_package_name "${use_package}" || return 1
	else
		set_pkg_portdir "${pkg_origin}"
		correct_pkg || return 1
		set_port_name "${pkg_portdir}" || return 1

		IFS='
'
		for i in ${MAKE_ARGS:+${MAKE_ARGS}}; do
			case ${pkg_origin} in
			${i%%=*}) opt_make_args="${opt_make_args} ${i#*=}" ;;
			esac
		done

		MAKE_CMD="${MAKE} ${opt_make_args}"
	fi
}

perform_replace() {
	local REPLACE_TMPDIR old_package old_required_by preserved_files

	if [ "${pkg_name##*-}" != "${port_name##*-}" ]; then
		info "Replacing '${pkg_name}' with '${port_name}'"
	elif is_yes ${opt_force}; then
		info "Reinstalling '${pkg_name}'"
	else
		is_yes ${opt_verbose} && warn "No need to replace '${pkg_name}'. (specify -f to force)"
		return 0
	fi

	if is_yes ${opt_noexecute}; then
		status=done
		return 0
	elif is_yes ${opt_interactive}; then
		prompt_yesno || return 0
	fi

	if [ -z "${use_package}" ]; then
		build_port "${pkg_portdir}" || { log="build error"; return 1; }
	fi

	REPLACE_TMPDIR=${PKG_TMPDIR}/${pkg_name}.XXXXXX

	if [ ${opt_debug} -lt 2 ]; then
		REPLACE_TMPDIR=$(mktemp -d "${REPLACE_TMPDIR}") || return 1
	fi

	old_package="${REPLACE_TMPDIR}/${pkg_name}${PKG_SUFX}"
	old_required_by="${REPLACE_TMPDIR}/+REQUIRED_BY"

	if ! {
		create_old_package "${pkg_name}" "${old_package}" && \
		create_old_required_by "${pkg_pkgdir}/+REQUIRED_BY" "${old_required_by}" && \
		preserve_shlibs "${pkg_name}"
	}; then
		log="backup error"
		cleanup "${REPLACE_TMPDIR}"
		return 1
	fi

	if opt_force=YES deinstall_package "${pkg_name}"; then
		if {
			if [ -n "${use_package}" ]; then
				install_package "${use_package}"
			else
				install_port "${pkg_portdir}" && \
				opt_use_index=NO set_port_name "${pkg_portdir}"
			fi
		}; then
			set_pkg_pkgdir "${port_name}"
			status=done
		else
			log="install error"
			restore_package "${old_package}" || return 1
		fi
	else
		log="deinstall error"
	fi

	restore_required_by "${old_required_by}" "${pkg_pkgdir}/+REQUIRED_BY" || return 1
	process_old_package "${old_package}" || :
	clean_shlibs || warn "Couldn't remove '${PKGCOMPATDIR}'."
	cleanup "${REPLACE_TMPDIR}"

	[ "${status}" = done ] || return 1

	if [ "${pkg_name}" != "${port_name}" ]; then
		info "Updating the dependencies"
		try update_dependencies "${pkg_name}" "${port_name}" || return 1
	fi
}

update_dependencies() {
	local oldname newname required_by dep

	oldname=$1
	newname=$2

	pkg_required_by "${newname}" '$required_by'
	for dep in ${required_by}; do
		have_depend "${dep}" "${oldname}" || continue
		update_contents "${dep}" "${oldname%-*}-[^-]*" "${newname}" || return 1
	done
}

update_contents() {
	local contents obj

	contents=${PKG_DBDIR}/$1/+CONTENTS
	obj=${contents}.$$

	if is_yes ${opt_verbose}; then
		info "Updating $1: $2 -> $3"
	fi

	if sed "s/^@pkgdep $2\$/@pkgdep $3/" "${contents}" > "${obj}"; then
		try mv -f "${obj}" "${contents}"
	else
		return 1
	fi
}

have_depend() {
	local dep

	if [ -r "${PKG_DBDIR}/$1/+CONTENTS" ]; then
		while read dep; do
			case ${dep} in
			@pkgdep\ ${2%-*}[\<\>*]*|@pkgdep\ ${2%-*}-*\**|[!@]*)
				break ;;
			@pkgdep\ ${2%-*}-[0-9]*)
				return 0 ;;
			esac
		done < "${PKG_DBDIR}/$1/+CONTENTS"
	fi

	return 1
}

preserve_shlibs() {
	local file

	is_yes ${opt_preserve_libs} || return 0

	preserved_files=
	for file in $(pkg_info -qL "$1"); do
		case ${file} in
		*.so.[0-9]*)
			case $(file -b "${file}") in
			*ELF*shared\ object*)
				preserved_files="${preserved_files} ${file}"
			esac ;;
		esac
	done

	if [ -n "${preserved_files}" ]; then
		info "Preserving the shared libraries"
		make_dirs "${PKGCOMPATDIR}" || return 1
		try cp -f ${preserved_files} "${PKGCOMPATDIR}" || return 1
		try ldconfig -m "${PKGCOMPATDIR}" || return 1
	fi
}

clean_shlibs() {
	local delete_files file dest

	if ! is_yes ${opt_preserve_libs} || [ -z "${preserved_files}" ]; then
		return 0
	fi

	info "Cleaning the preserved shared libraries"

	delete_files=
	for file in ${preserved_files}; do
		dest=${PKGCOMPATDIR}/${file##*/}
		if [ -e "${file}" ]; then
			delete_files="${delete_files} ${dest}"
		else
			info "Keeping ${file} as ${dest}"
		fi
	done

	if [ -n "${delete_files}" ]; then
		try rm -f ${delete_files} || return 1
		try ldconfig -m "${PKGCOMPATDIR}" || return 1
	fi
}

file_exist() {
	if [ ! -e "$1" ]; then
		warn "No such file or directory: $1"
		return 1
	fi
}

make_dirs() {
	while [ $# -gt 0 ]; do
		[ -d "$1" ] || try mkdir -p "$1" || return 1
		shift
	done
}

parse_args() {
	local port_name file

	while [ $# -gt 0 ]; do
		case $1 in
		*.t[bg]z|?*=${PORTSDIR}/?*/*[!/])
			file=${1##*=}
			expand_path '$file'
			file_exist "${file}" || return 1 ;;
		*=*)
			warn "Syntax error: $1"
			return 1 ;;
		esac

		case $1 in
		*=*)
			opt_replace="${opt_replace} ${1%=*}=${file}"
			do_replace="${do_replace} ${1%=*}" ;;
		*.t[bg]z)
			set_package_name "${file}" || return 1
			opt_replace="${opt_replace} ${port_name%-*}=${file}"
			do_replace="${do_replace} ${port_name%-*}" ;;
		*)
			do_replace="${do_replace} $1" ;;
		esac
		shift
	done
}

create_log() {
	[ -n "$1" ] || return 0

	log_output=$1
	expand_path '$log_output'

	true > "${log_output}" || {
		warn "Failed to create the log."
		return 1; }
}

add_log() {
	local sign valid

	[ -n "${log_output}" ] || return 0

	valid=
	for sign in ${log_format}; do
		case $1 in
		${sign#*:})
			valid=${sign%%:*}
			break ;;
		esac
	done; shift

	echo "${valid:-?} $@" >> "${log_output}"
}

show_log() {
	local LINE i descr msg

	[ -n "${log_output}" ] || return 0

	all_c=0
	descr=
	msg=

	for i in ${log_format}; do
		eval ${i#*:}_c=0
		descr="${descr:+${descr} / }$i"
	done

	info "Listing the results (${descr})"

	while read LINE; do
		all_c=$((${all_c}+1))
		for i in ${log_format}; do
			case ${LINE} in
			"${i%%:*} "*)
				eval ${i#*:}_c=\$\(\(\${${i#*:}_c}+1\)\)
				break ;;
			esac
		done
		echo "        ${LINE}"
	done < "${log_output}"

	for i in ${log_format}; do
		eval msg=\"${msg:+${msg}, }\${${i#*:}_c} ${i#*:}\"
	done

	info "All tasks ${all_c}: ${msg%,*} and ${msg##*, }"
	info "Saved the results to '${log_output}'"
}

expand_path() {
	local val xvar

	case $1 in
	\$*)	xvar=${1#$}; eval val=$1 ;;
	*)	xvar=; val=$1 ;;
	esac

	case ${val} in
	[!/]*)	val=${cwd%/}/${val}
	esac
	
	if [ -n "${xvar}" ]; then
		eval ${xvar}=\${val}
	else
		echo "${val}"
	fi 
}

main() {
	init_variables
	init_options
	parse_options ${1+"$@"}
	shift ${OPTC}

	if is_yes ${opt_all}; then
		opt_depends=NO
		opt_required_by=NO
		set -- '*'
	elif [ $# -eq 0 ]; then
		usage
	fi

#ifdef WITH_EXPERIMENTAL
	if ! is_yes ${opt_noconf}; then
		if ! parse_config "/usr/local/etc/pkg_replace.conf"; then
			warn "Fatal error in /usr/local/etc/pkg_replace.conf."
			exit 1
		fi
	fi
#endif
	compat_variables

	if is_yes ${opt_use_index} && ! file_exist "${PORTSDIR}/INDEX" ; then
		warn "Couldn't use -I option."
		usage 1
	fi

	parse_args ${1+"$@"} || usage 1
	set -- $(pkg_glob ${do_replace})

	if [ $# -gt 1 ]; then
#ifdef WITH_PORTS
		if is_yes ${opt_use_index} && ! is_yes ${opt_force}; then
			guess_up_to_date "$@"
			set -- ${do_replace}

			if [ $# -eq 0 ]; then
				warn "No need to replace packages. (specify -f to force)"
				exit 0
			fi
		fi
#endif
		set -- $(pkg_sort "$@")
	fi

	if [ ${opt_debug} -eq 1 ]; then
		while [ $# -gt 0 ]; do
			echo "${0##*/} $1"
			shift
		done
	elif [ $# -gt 0 ]; then
		do_replace=
		create_log "${opt_log_output}" || usage 1

		perform ${1+"$@"}

		show_log
	fi

	exit 0
}

IFS=' 
'
main ${1+"$@"}
