#! /bin/sh -efu
# $Id: pkgdb_fix.sh,v 1.1 2006/10/15 19:52:16 securedog Exp $

usage() {
	echo "usage: ${0##*/} [-habBfFiQrRv] [-s /OLD/NEW/] pkgname[=origin] ..."
	exit ${1:-0}
}

is_yes() {
	case ${1-} in
	[Yy][Ee][Ss]|[Yy]|[Tt][Rr][Uu][Ee])
		return 0 ;;
	*)	return 1 ;;
	esac
}

warn() {
	echo "** $@" >&2
}

prompt_yesno() {
	local prompt default input

	prompt=${1-"OK?"}
	default=${2-"yes"}

	echo -n "${prompt} [${default}] " >&2
	read input

	is_yes ${input:-"${default}"}
}

init_variables() {
#ifdef WITH_PORTS
	: ${PORTSDIR="/usr/ports"}
#endif
#ifdef WITH_PKGSRC
	: ${PKGSRCDIR="/usr/pkgsrc"}
#endif
	: ${PKG_DBDIR="/var/db/pkg"}
	: ${TMPDIR="/tmp"}
	REGENERATE_REQUIRED_BY_COOKIE=.regenerate.required_by
#ifdef WITH_PORTS
	export PORTSDIR PKG_DBDIR
#endif
#ifdef WITH_PKGSRC
	PORTSDIR=${PKGSRCDIR}
	export PKGSRCDIR PKG_DBDIR
#endif
}

init_options() {
	opt_all=NO
	opt_update_reqby=YES
	opt_force=NO
	opt_fix=NO
	opt_fix_duporigin=NO
	opt_interactive=NO
	opt_quiet=NO
	opt_substitute=
	opt_verbose=NO
	opt_depends=NO
	opt_required_by=NO
	opt_replace=
	opt_origin=
	packages=
}

parse_options() {
	local OPTS OPT OPTARG OPTIND

	while getopts habBfFhiQrRs:v OPT; do
		case ${OPT} in
		a)	opt_all=YES ;;
		b)	opt_update_reqby=ALWAYS ;;
		B)	opt_update_reqby=NO ;;
		f)	opt_force=YES ;;
		F)	opt_fix=YES ;;
		h)	usage ;;
		i)	opt_interactive=YES ;;
		Q)	opt_quiet=YES ;;
		r)	opt_required_by=YES ;;
		R)	opt_depends=YES ;;
		s)	opt_substitute="${opt_substitute} ${OPTARG}" ;;
		v)	opt_verbose=YES ;;
		\?)	usage 1 ;;
		esac
	done
	OPTC=$((${OPTIND}-1))
}

bre2regex() {
	echo "$@" |  sed -e 's#\([?^$.|/(){}]\)#\\\1#g' -e 's,*,.*,g'
}

regex_quote() {
	echo "$@" |  sed -e 's#\([][?^$.*|/(){}]\)#\\\1#g'
}

is_installed() {
#ifdef WITH_PORTS
	if [ -e "${PKG_DBDIR}/$1/+CONTENTS" ]; then
#endif
#ifdef WITH_PKGSRC
	if pkg_info -qe "$1"; then
#endif
		return 0
	else
		return 1
	fi
}

pkg_glob() {
	local - pattern name contents

	set +f
	while [ $# -gt 0 ]; do
		pattern=${1#${PKG_DBDIR}/}

		case ${pattern} in
		''|*[\<\>/]*)	shift; continue ;;
		*\**|*-pl[0-9]*|*-[0-9]*[0-9.][a-z]|*-[0-9]*[0-9]) ;;
		*)		pattern="${pattern}-[0-9]*[0-9a-z]" ;;
		esac

		for contents in ${PKG_DBDIR}/${pattern}/+CONTENTS; do
			if [ -e "${contents}" ]; then
				name=${contents%/+CONTENTS}

				if is_yes ${opt_depends}; then
					pkg_depends "${name##*/}"
				fi

				echo "${name##*/}"

				if is_yes ${opt_required_by}; then
					pkg_required_by "${name##*/}"
				fi
			else
				warn "No such installed package: $1"
			fi
		done
		shift
	done
}

pkg_depends() {
	local dep pkg xvar

	case ${2-} in
	\$*)	xvar=${2#$}; eval ${xvar}= ;;
	*)	xvar= ;;
	esac

	if [ -r "${PKG_DBDIR}/$1/+CONTENTS" ]; then
		while read dep; do
			case ${dep} in
#ifdef WITH_PORTS
			@pkgdep\ *)
				if [ -n "${xvar}" ]; then
					eval ${xvar}=\"\$${xvar} ${dep#@pkgdep }\"
				else
					echo "${dep#@pkgdep }"
				fi ;;
#endif
#ifdef WITH_PKGSRC
			@pkgdep\ *)
				case ${dep#@pkgdep } in
				*[\<\>{}*]*)	pkg=$(pkg_info -e "${dep#@pkgdep }") ;;
				*)		pkg=${dep#@pkgdep } ;;
				esac

				if [ -n "${xvar}" ]; then
					eval ${xvar}=\"\$${xvar} ${pkg}\"
				else
					echo "${pkg}"
				fi ;;
#endif
			[!@]*)	break ;;
			esac

		done < "${PKG_DBDIR}/$1/+CONTENTS"
	fi
}

pkg_required_by() {
	local req xvar
	
	case ${2-} in
	\$*)	xvar=${2#$}; eval ${xvar}= ;;
	*)	xvar= ;;
	esac

	if [ -r "${PKG_DBDIR}/$1/+REQUIRED_BY" ]; then
		while read req; do
			if [ -n "${xvar}" ]; then
				eval ${xvar}=\"\$${xvar} ${req}\"
			else
				echo "${req}"
			fi
		done < "${PKG_DBDIR}/$1/+REQUIRED_BY"
	fi
}

pkg_required_by_regex() {
	local - req
	set +f

	echo ${PKG_DBDIR}/*/+CONTENTS | xargs grep -l "^@pkgdep ${1}$" | \
	while read req; do
		req=${req%/+CONTENTS}
		echo "${req##*/}"
	done
}

set_pkg_name() {
	pkg_name=$1
}

set_pkg_pkgdir() {
	pkg_pkgdir=${PKG_DBDIR}/$1
}

set_pkg_origin() {
#ifdef WITH_PORTS
	local LINE

	pkg_origin=
	if [ -r "${PKG_DBDIR}/$1/+CONTENTS" ]; then
		while read LINE; do
			case ${LINE} in
			@comment\ ORIGIN:*)
				pkg_origin=${LINE#@comment ORIGIN:}
				break ;;
			[!@]*)	break ;;
			esac
		done < "${PKG_DBDIR}/$1/+CONTENTS"
	fi
#endif
#ifdef WITH_PKGSRC
	pkg_origin=$(pkg_info -Q PKGPATH "$1")
#endif
}

create_pkg_copy() {
	local source

	source=${PKG_DBDIR}/$1/$2
	created=${WORKDIR}/$1$2

	if [ -e "${source}" ]; then
		cp "${source}" "${created}" || return 1
	else
		: > "${created}" || return 1
	fi
}

set_pkg_contents() {
	if [ -z "${pkg_contents}" ]; then
		create_pkg_copy "$1" "+CONTENTS" || return 1
		pkg_contents=${created}
	fi
}

set_pkg_build_info() {
	if [ -z "${pkg_build_info}" ]; then
		create_pkg_copy "$1" "+BUILD_INFO" || return 1
		pkg_build_info=${created}
	fi
}

pkg_preprocess() {
	local replace

	pkg_contents=
	pkg_build_info=
	set_pkg_name "$1"
	set_pkg_pkgdir "$1"

	if [ ! -e "${pkg_pkgdir}/+CONTENTS" ]; then
		warn "'${pkg_name}' is not installed. (broken dependencies?)"
		return 1
	fi

	set_pkg_origin "$1"
	set_pkg_contents "$1"

	opt_origin=
	for replace in ${opt_replace}; do
		case ${pkg_name} in
		${replace%=*}|${replace%=*}-[0-9]*)
			opt_origin="${replace##*=}"
			break ;;
		esac
	done
}


run_pkgdb_fix() {
	local -	pkg_name pkg_pkgdir pkg_origin pkg_contents pkg_build_info \
		_subst _subst_old _subst_new

	pkg_preprocess "$1" || return 0

	if [ -z "${opt_origin}${opt_substitute}" ]; then
		is_yes ${opt_verbose} && echo "Checking for ${pkg_name}."

		pkgdb_fix_origin || return 0
		pkgdb_fix_dependencies || return 0

		if is_yes ${opt_update_reqby} && [ ! -e "${pkg_pkgdir}/+REQUIRED_BY" ] \
			|| [ ${opt_update_reqby} = ALWAYS ]; then
			echo "${pkg_name}" >> "${WORKDIR}/${REGENERATE_REQUIRED_BY_COOKIE}"
		fi
	else
		if [ -n "${opt_origin}" ]; then
			is_yes ${opt_verbose} && echo "${pkg_name}: ${pkg_origin} -> ${opt_origin}"
			modify_origin "${opt_origin}"
		fi

		for _subst in ${opt_substitute}; do
			case ${_subst} in
			*/*/*/*/*)
				warn "Illegal expression: ${_subst}" ;;
			/?*/*/)
				_subst=${_subst#/}
				_subst=${_subst%/}
				_subst_old=${_subst%/*}
				_subst_new=${_subst#*/}

				if [ -n "${_subst_new}" ]; then
					is_yes ${opt_verbose} && \
					echo "${pkg_name}: replacing dependency: ${_subst_old} -> ${_subst_new}"
					modify_dep "${_subst_old}" "${_subst_new}"
				else
					is_yes ${opt_verbose} && \
					echo "${pkg_name}: deleting dependency: ${_subst_old}"
					delete_dep "${_subst_old}"
				fi ;;
			*)
				warn "Illegal expression: ${_subst}" ;;
			esac
		done
	fi
}

run_pkgdb_fix_duporigin() {
	local pkg_origin

	is_yes ${opt_verbose} && echo "Checking for origin duplicates."

#ifdef WITH_PORTS
	for pkg_origin in $(pkg_info -aqo | sort | uniq -d); do
#endif
#ifdef WITH_PKGSRC
	for pkg_origin in $(pkg_info -aQ PKGPATH | sort | uniq -d); do
#endif
		pkgdb_fix_duporigin "${pkg_origin}"
	done
}

replace() {
	sed "$1" "$2" > "$2.new"
	mv -f "$2.new" "$2"
}

modify_origin() {
        local new_origin

        new_origin=$1

#ifdef WITH_PORTS
	if grep -q "^@comment ORIGIN:" ${pkg_contents}; then
		replace "s|^@comment ORIGIN:.*$|@comment ORIGIN:${new_origin}|" "${pkg_contents}"
	else
		replace "/^@name /a\\
@comment ORIGIN:${new_origin}
" "${pkg_contents}"
	fi
#endif
#ifdef WITH_PKGSRC
	set_pkg_build_info "${pkg_name}"

	if grep -q "^PKGPATH=" ${pkg_build_info}; then
		replace "s|^PKGPATH=.*$|PKGPATH=${new_origin}|" "${pkg_build_info}"
	else
		echo "PKGPATH=${new_origin}" >> "${pkg_build_info}"
	fi
#endif
}

delete_dep() {
	local deppkg

	deppkg=$(bre2regex "$1")
	grep -q "^@pkgdep ${deppkg}$" "${pkg_contents}" || return 0

#ifdef WITH_PORTS
	replace "/^@pkgdep ${deppkg}$/,/^@comment DEPORIGIN:/d" "${pkg_contents}"
#endif
#ifdef WITH_PKGSRC
	replace "/^@blddep /,/^@pkgdep ${deppkg}$/d" "${pkg_contents}"
#endif
}

modify_dep() {
	local oldpkg newpkg neworigin

	oldpkg=$(bre2regex "$1")
	newpkg=$2
	neworigin=${3-}

	grep -q "^@pkgdep ${oldpkg}$" "${pkg_contents}" || return 0

#ifdef WITH_PORTS
	if [ -n "${neworigin}" ]; then
		replace "/^@pkgdep ${oldpkg}$/,/^@comment DEPORIGIN:/c\\
@pkgdep ${newpkg}\\
@comment DEPORIGIN:${neworigin}
" "${pkg_contents}"
	else
#endif
		replace "s|^@pkgdep ${oldpkg}$|@pkgdep ${newpkg}|g" "${pkg_contents}"
#ifdef WITH_PORTS
	fi
#endif
}

trace_moved() {
#ifdef WITH_PORTS
	local IFS msg checked moved origin

	if [ ! -e "${PORTSDIR}/MOVED" ]; then
		warn "'${PORTSDIR}/MOVED' does not exist."
		return 1
	fi

	msg=
	origin=$1
	checked=
	IFS='|'

	while moved=$(grep -m1 "^$1|" "${PORTSDIR}/MOVED"); do
		set -- ${moved}

		if [ -n "$2" ]; then
			checked="${checked:-|}$1|"

			case ${checked} in
			*"|${2}|"*)
				warn "Failed. (${msg} -> ???)"
				break ;;
			esac

			msg="${msg:+${msg} }-> $2"

			if [ -d "${PORTSDIR}/$2" ]; then
				warn "'${origin}' has been moved. (${msg})"
				pkg_origin=$2
				return 0
			fi

			set -- "$2"
		else
			warn "'$1' has been removed from ports tree:"
			warn "\"$4\" ($3)"
			return 1
		fi
	done

	warn "Port directory not found: ${origin}"
	return 1
#endif
#ifdef WITH_PKGSRC
	warn "See ${PORTSDIR}/doc/CHANGES-*."
	return 1
#endif
}

pkgdb_fix_origin() {
	if [ -z "${pkg_origin}" ]; then
		warn "'${pkg_name}' has no origin recorded."
		echo "Skipped."
		return 1
	elif [ ! -d "${PORTSDIR}/${pkg_origin}" ]; then
		echo "Stale origin: ${pkg_name} (${pkg_origin}): perhaps moved or obsoleted."

		if trace_moved "${pkg_origin}"; then
			modify_origin "${pkg_origin}"
			echo "Fixed -> (${pkg_origin})"
		elif is_yes ${opt_fix} && prompt_yesno "Delete this?" "no"; then
			pkg_delete -f "${pkg_name}"
			return 1
		else
			echo "Skipped."
			return 1
		fi
	fi
}

fix_dep() {
#ifdef WITH_PORTS
	local pkg_origin

	if [ -n "${3:+fix_origin}" ]; then
		set_pkg_origin "$2"
		modify_dep "$1" "$2" "${pkg_origin}"
		echo "Fixed. (-> $2, ${pkg_origin})"
	else
#endif
		modify_dep "$1" "$2"
		echo "Fixed. (->$2)"
#ifdef WITH_PORTS
	fi
#endif
}

fix_multiple_dep() {
	local oldpkg pkg_origin

	oldpkg=$1; shift
	while [ $# -gt 0 ]; do
		set_pkg_origin "$1"
		if prompt_yesno "Use $1 (${pkg_origin})?"; then
#ifdef WITH_PORTS
			modify_dep "${oldpkg}" "$1" "${pkg_origin}"
			echo "Fixed. (-> $1, ${pkg_origin})"
#endif
#ifdef WITH_PKGSRC
			modify_dep "${oldpkg}" "$1"
			echo "Fixed. (-> $1)"
#endif
			return 0
		fi
		shift
	done

	return 1
}

try_fix_dep() {
	local - input pkg found

	set +f
	while :; do
		echo -n "New dependency? " >&2
		read input

		case ${input} in
		''|[Nn][Oo]|[Nn])
			if prompt_yesno "Delete this?" "no"; then
				delete_dep "$1"
				echo "Deleted."
				break
			elif prompt_yesno "Skip this?" "yes"; then
				break
			fi ;;
		*)
			if is_installed "${input}"; then
				fix_dep "$1" "${input}" 1
				break
			else
				found=
				for pkg in ${PKG_DBDIR}/${input%-*}*; do
					case ${pkg} in
					*\**) ;;
					*) found="${found} ${pkg##*/}" ;;
					esac
				done

				if [ -n "${found}" ]; then
					echo "Please choose one of these:" >&2
					echo "${found}" >&2
				else
					echo "'${input}' is not installed." >&2
				fi
			fi
		esac
	done
}

pkgdb_fix_duporigin() {
	local pkg pkg_delete

#ifdef WITH_PORTS
	set -- $(pkg_info -qO "${pkg_origin}")
#endif
#ifdef WITH_PKGSRC
	set -- $(echo ${PKG_DBDIR}/*/+BUILD_INFO | xargs grep -l "^PKGPATH=${pkg_origin}")
#endif
	pkg_delete=

	if [ $# -gt 1 ]; then
		echo "Duplicated origin: ${pkg_origin} - $@" >&2
		echo "Please remove unnecessary packages."

		while [ $# -gt 1 ]; do
			if prompt_yesno "Delete $1?" "no"; then
				pkg_delete="${pkg_delete} $1"
			else
				set -- "$@" "$1"
			fi
			shift
		done

		for pkg in ${pkg_delete}; do
			echo "Unregistering ${pkg}." >&2
			rm -rf "${PKG_DBDIR}/${pkg}"
		done
	fi
}

pkgdb_fix_dependencies() {
	local LINE dep_pkg_name dep_pkg_origin pkg_origin

	IFS='
'
#ifdef WITH_PORTS
	for LINE in $(grep -e '^@pkgdep ' -e '^@comment DEPORIGIN:' "${PKG_DBDIR}/${pkg_name}/+CONTENTS"); do
		case ${LINE} in
		@pkgdep\ *) dep_pkg_name=${LINE#@pkgdep }; continue ;;
		@comment\ DEPORIGIN:*) dep_pkg_origin=${LINE#@comment DEPORIGIN:} ;;
		esac
#endif
#ifdef WITH_PKGSRC
	for LINE in $(grep -e '^@pkgdep' "${PKG_DBDIR}/${pkg_name}/+CONTENTS"); do
		dep_pkg_name=${LINE#@pkgdep }
#endif

		if ! is_installed "${dep_pkg_name}"; then
#ifdef WITH_PORTS
			echo "Stale dependency: ${pkg_name} -> ${dep_pkg_name} (${dep_pkg_origin}):"
			set -- $(pkg_info -qO "${dep_pkg_origin}")

#endif
#ifdef WITH_PKGSRC
			echo "Stale dependency: ${pkg_name} -> ${dep_pkg_name}:"

			set -- $(pkg_info -e "${dep_pkg_name%[-<>]*}*")
#endif

			if [ $# -eq 1 ]; then
				fix_dep "${dep_pkg_name}" "$1"
				continue
#ifdef WITH_PORTS
			elif [ $# -gt 1 ]; then
				warn "Detected the origin duplicates, please manually run '${0##*/} -aF' to fix."
				return 1
			fi

			set -- $(pkg_info -E "${dep_pkg_name%-*}*" "??-${dep_pkg_name%-*}*")

			if [ $# -eq 1 ]; then
				fix_dep "${dep_pkg_name}" "$1" 1
				continue
#endif
			elif [ $# -gt 1 ] && is_yes ${opt_fix}; then
				echo "Detected the following packages:" >&2
				echo "$@" >&2

				if fix_multiple_dep "${dep_pkg_name}" "$@"; then
					continue
				fi
			fi

			if is_yes ${opt_fix}; then
				try_fix_dep "${dep_pkg_name}"
			else
				echo "Skipped. (specify -F to fix)"
			fi
		fi
	done
}

regenerate_required_by() {
	local pkg_name

	[ -e "${WORKDIR}/${REGENERATE_REQUIRED_BY_COOKIE}" ] || return 0

	is_yes ${opt_verbose} && echo "Regenerating +REQUIRED_BY files."

	while read pkg_name; do
		pkg_required_by_regex "${pkg_name}" > "${WORKDIR}/${pkg_name}+REQUIRED_BY"
	done < "${WORKDIR}/${REGENERATE_REQUIRED_BY_COOKIE}"

	rm -f "${WORKDIR}/${REGENERATE_REQUIRED_BY_COOKIE}"
}

pkgdb_fix_update() {
	local - source target file

	set +f

	for source in ${WORKDIR}/*+*; do
		case ${source} in
		*\**) break ;;
		esac

		file=${source##*/}
		target="${PKG_DBDIR}/${file%+*}/+${file##*+}"

		if confirm_install "${source}" "${target}"; then
			cp -f "${source}" "${target}" || :
		fi
	done

	rm -f ${WORKDIR}/*+* || :
}

confirm_install() {
	local old_file new_file

	new_file=$1
	old_file=$2

	[ -s "${new_file}" ] || return 1

	if [ -e "${old_file}" ]; then
		if ! cmp -s "${old_file}" "${new_file}"; then
			if is_yes ${opt_interactive}; then
				diff -u "${old_file}" "${new_file}" >&2
				prompt_yesno "Install?" && return 0
			else
				return 0
			fi
		fi
	elif is_yes ${opt_interactive}; then
		cat "${new_file}" >&2
		prompt_yesno "Install ${old_file}?" && return 0
	else
		return 0
	fi

	return 1
}

pkgdb_fix_clean() {
	rm -rf "${WORKDIR}"
}

pkgdb_fix_done() {
	pkgdb_fix_update
	regenerate_required_by
	pkgdb_fix_update
	pkgdb_fix_clean
}

check_permission() {
	local target

	target=$1
	if [ ! -w "${target}" ]; then
		if is_yes ${opt_force}; then
			warn "You do not own ${target}. (proceeding anyway)"
		else
			warn "You do not own ${target}. (specify -f to force)"
			return 1
		fi
	fi
}

create_working_dir() {
	WORKDIR=$(mktemp -dt .pkgdb) || return 1
	trap 'pkgdb_fix_clean; exit' 1 2 3 15
}

parse_args() {
	while [ $# -gt 0 ]; do
		case $1 in
		?*=?*)
			opt_replace="${opt_replace} ${1}"
			packages="${packages} ${1%=*}" ;;
		*=*)
			warn "Syntax error: $1"
			return 1 ;;
		*)
			packages="${packages} $1" ;;
		esac
		shift
	done
}

main() {
	init_variables
	init_options

	parse_options ${1+"$@"}
	shift ${OPTC}

	is_yes ${opt_quiet} && exec >/dev/null

	if is_yes ${opt_all}; then
		opt_depends=NO
		opt_required_by=NO
		is_yes ${opt_fix} && opt_fix_duporigin=YES
		set -- '*'
	fi

	[ $# -gt 0 ] || usage

	check_permission "${PKG_DBDIR}" || return 1
	create_working_dir || return 1

	is_yes ${opt_verbose} && echo "Checking the package registry database."
	is_yes ${opt_fix_duporigin} && run_pkgdb_fix_duporigin

	parse_args ${1+"$@"}
	set -- $(pkg_glob ${packages})

	while [ $# -gt 0 ]; do
		run_pkgdb_fix "$1"
		shift
	done

	pkgdb_fix_done
}

IFS=' 
'
main ${1+"$@"}
